#ProcessorDeposition.py
'''Functions to process the deposition component.'''

import logging
import os
from glob import glob
from pathlib import Path
from string import Template

import iris
from iris.cube import CubeList

from ews.coordinator.utils import processor_utils
from ews.coordinator.utils.processor_utils import subprocess_and_log, get_only_existing_globs
from ews.coordinator.processor_base import ProcessorBase
from ews.postprocessing.deposition.deposition_post_processor import DepositionPostProcessor

logger = logging.getLogger(__name__)

class ProcessorDeposition(ProcessorBase):

    """ LIFECYCLE FUNCTIONS INHERITED FROM PROCESSOR.PY """

    def process_pre_job(self, args) -> bool:
        # return process_pre_job_server_download(args)
        return True


    def process_in_job(self, jobPath, status, configjson, component) -> dict:
        return self.process_in_job_dep(jobPath, status, configjson, component)


    def process_post_job(self, jobPath, configjson):
        self.process_EWS_plotting_dep(jobPath, configjson)


    def __init__(self) -> None:
        super().__init__()
        print(__name__)

    """ LIFECYCLE FUNCTIONS INHERITED FROM PROCESSOR.PY """


    def process_in_job_dep(self, jobPath, status, config, component):
        logger.info('started process_in_job_dep()')

        file_path = Template(config[component]['ServerPathTemplate']).substitute(**config)
        file_name = Template(config[component]['InputFileTemplate']).substitute(**config)

        logger.info(f"Expecting to work with {file_name}")

        if os.path.exists(f"{jobPath}/{file_name}"):
            logger.info('Directory already exists in job directory, so nothing to do here')
            return

        logger.info('Copying file from remote server to job directory')

        # TODO: perform ssh file transfer in python instead of subprocess
        server_name: str = config['ServerName']
        if server_name == "":
            cmd_scp = ["scp", f"{file_path}/{file_name}.tar.gz", jobPath]
        else:
            cmd_scp = ["scp", "-i", config['ServerKey'], "-o", "StrictHostKeyChecking=no",
                       f"{server_name}:{file_path}/{file_name}.tar.gz", jobPath]

        description_short = 'dep scp'
        description_long = 'scp from server to job directory'
        subprocess_and_log(cmd_scp, description_short, description_long)

        logger.info('untarring the input file')

        # TODO: untar file in python (with tarfile module) instead of subprocess
        cmd_tar = ["tar","-xzf",f"{jobPath}/{file_name}.tar.gz","-C",jobPath]
        description_short = 'dep tars'
        description_long = 'untar the downloaded file'
        subprocess_and_log(cmd_tar, description_short, description_long)

        # basic check that contents are as expected
        # 132 files of NAME .txt timesteps and one summary png file
        # if len(glob(f"{jobPath}/{file_name}/deposition_srcs_allregions_C1_T*.txt")) != 56:
        #     msg = f"Unexpect number of deposition .txt files in input tar file. Expected 56."
        #     logger.error(msg)
        #     raise RuntimeError(msg)

        # basic check that contents are as expected (56 timepoints in the file)
        cube_wildcard = f"{jobPath}/{file_name}/deposition_srcs_allregions*.nc"
        cubes: CubeList = iris.load(cube_wildcard)
        for cube in cubes:
            coord = cube.coord("time")
            timepoint_count = coord.shape[0]
            if timepoint_count != 56:
                msg = f"Unexpected number of timepoints ({timepoint_count}) in cube {cube.name()}"
                logger.error(msg)
                raise RuntimeError(msg)

        proc_out = {}
        # Output files available for upload
        proc_out['output'] = None
        # Processing files available for clearing
        proc_out['clearup'] = [f"{jobPath}/{file_name}.tar.gz"]

        return proc_out


    def process_EWS_plotting_dep(self, jobPath, config) -> [str]:
        '''Returns a list of output files for transfer.'''

        logger.info('started process_EWS_plotting_dep()')

        # initialise environment
        regions = config['SubRegionNames']

        deposition_file_name = Template(config['Deposition']['InputFileTemplate']).substitute(**config)

        deposition_path = f"{jobPath}/{deposition_file_name}"

        # get the file name from the config
        # this file name can be a glob, as long as matches can all be loaded by iris
        deposition_data_file_name = Template(config['Deposition']['DataFileTemplate']).substitute(**config)
        name_file_wildcard = f"{deposition_path}/{deposition_data_file_name}"

        ews_plotting_output_globs = []

        for region in regions:

            output_dir = f"{jobPath}/plotting/{region.lower()}"

            Path(output_dir).mkdir(parents=True, exist_ok=True)

            sys_config = config['Deposition']['EWS-Plotting']['SysConfig']
            name_extraction_config = config['Deposition']['EWS-Plotting']['NameExtractionConfig']
            run_config = config['Deposition']['EWS-Plotting']['RunConfig']
            run_config_norm = config['Deposition']['EWS-Plotting']['RunConfigNorm']
            chart_config = config['Deposition']['EWS-Plotting'][region]['ChartConfig']
            normalize = config['Deposition']['EWS-Plotting'][region]['Normalize']
            extraction_file_prefix = 'deposition_' + region.lower()

            # Note that this runs all disease types available

            logger.info(f"Running EWS-Plotting with the following configs:\n{sys_config}\n{name_extraction_config}\n{run_config}\n{run_config_norm}\n{chart_config}")

            depo_processor = DepositionPostProcessor()
            depo_processor.set_param_config_files(sys_config_file_arg = sys_config,
                                                  depo_name_extraction_config_file_arg = name_extraction_config,
                                                  chart_config_file_arg = chart_config,
                                                  depo_plotting_run_config_file_arg = run_config,
                                                  depo_plotting_normalized_run_config_file_arg = run_config_norm,
                                                  name_file_wildcard_arg = name_file_wildcard,
                                                  wheat_sources_dir_arg = deposition_path,
                                                  output_dir_arg = output_dir,
                                                  issue_date_arg = config['StartString'],
                                                  extraction_file_prefix_arg = extraction_file_prefix)


            # asia/east africa env suit should not perform normalization, false gets passed here for these areas
            depo_processor.name_extract_params.NORMALIZE = (normalize.upper() == "TRUE")

            depo_processor.process()

            # check the output
            ews_plotting_output_dir = f"{output_dir}/images/"
            ews_plotting_output_globs += [f"{ews_plotting_output_dir}*"]

        ews_plotting_output_globs = get_only_existing_globs(ews_plotting_output_globs, inplace = False)

        # check there is some output from EWS-plotting
        if not ews_plotting_output_globs:
            logger.error('EWS-Plotting did not produce any output')
            raise RuntimeError

        # provide list for transfer
        ews_plotting_outputs: [str] = sorted([file for glob_str in ews_plotting_output_globs for file in glob(glob_str)])

        return ews_plotting_outputs


if __name__ == '__main__':
    print(__name__)
    print(processor_utils.__name__)
    processor = ProcessorDeposition()
    processor.run_processor("Deposition", "DEPOSITION")

