#ProcessorEpidemiology.py
'''Functions to process the epidemiology component.'''

import datetime
from glob import glob
import json
import logging
import os
from pathlib import Path
import shutil
from typing import List

from numpy import all, any, argmax, unique, allclose
from pandas import MultiIndex, read_csv, DataFrame, to_datetime
from rasterio import open as rio_open

# gitlab projects
# TODO: Package these projects so they are robust for importing
from EpiModel import ( # created by rs481
    EpiAnalysis,
    model,
    plotRaster
)
from EpiModel.EpiPrep import prep
from ews.coordinator.processor_base import ProcessorBase
from ews.postprocessing.epi.epi_post_processor import EPIPostPostProcessor

from ews.coordinator.utils.processor_utils import (
    calc_epi_date_range,
    open_and_check_config,
    get_only_existing_globs,
    end_job,
    disease_latin_name_dict, short_name
)
logger = logging.getLogger(__name__)

class ProcessorEpidemiology(ProcessorBase):

    def process_pre_job(self, args) -> bool:
        return self.process_pre_job_epi(args)


    def process_in_job(self, jobPath, status, configjson, component) -> dict:
        return self.process_in_job_epi(jobPath, status, configjson, component)


    def process_post_job(self, jobPath, configjson):
        self.process_EWS_plotting_epi(jobPath, configjson)


    def __init__(self) -> None:
        super().__init__()

    def process_pre_job_epi(self, input_args: dict):
        '''Returns a boolean as to whether the job is ready for full processing.'''

        logger.info('started process_pre_job_epi()')

        # check pre-requisite jobs are complete
        # query_past_successes(input_args, 'Deposition')
        #
        # query_past_successes(input_args, 'Environment')

        config_fns: List[str] = input_args['config_paths']

        for configFile in config_fns:

            # they should be working if the script made it this far, no need to try
            config_i = open_and_check_config(configFile)

            #determine end time, from config file
            arg_start_date: str = input_args['start_date']
            calc_span_days = config_i['Epidemiology']['CalculationSpanDays']
            assert len(calc_span_days) == 2

            start_time, end_time = calc_epi_date_range(arg_start_date,calc_span_days)

            # warn if it is a long timespan
            date_diff = end_time - start_time
            if date_diff.days > 100:
                logger.warning("More than 100 days will be calculated over, likely longer than any single season")

        return True


    def create_epi_config_string(self, config,jobPath,startString,endString):

        configtemplate_fn = config['ConfigFilePath']
        configName_withoutEpi = f"{os.path.basename(configtemplate_fn).replace('.json','')}_{startString}-{endString}"

        # create a string describing every epi calc configuration
        epiStrings = []
        for epiconf in config['Epidemiology']['Epi']:
            epiKwargsString = ''.join([f"{k}{v}" for k,v in epiconf['modelArguments'].items()])

            # drop any repetitive elements of kwarg
            epiKwargsString = epiKwargsString.replace('infectionprevious','')
            epiKwargsString = epiKwargsString.replace('capbeta','cb')

            epiCaseString = f"{epiconf['model'].lower()}{epiKwargsString}"

            # provide to configuration for output filename
            epiconf["infectionRasterFileName"] = f"{jobPath}/infections_{configName_withoutEpi}_{epiCaseString}"

            epiStrings += [epiCaseString]

        epiString = '-'.join(epiStrings)

        config_filename = f"{configName_withoutEpi}_{epiString}"

        logger.debug(f"length of config filename is {len(config_filename)}.")

        if len(config_filename) > 254:
            logger.info(f"filename length is too long, it will raise an OSError, using a short form instead")

            # epi cases are not described in filename, an interested user
            # must look in the json file for details.
            config_filename = configName_withoutEpi

            assert len(config_filename) <= 254

        return config_filename

    def are_indices_close(self, idx1: MultiIndex, idx2: MultiIndex, atol=2.51e-6) -> bool:
        """An absolute tolerance of 2.51e-6 relates to differences between the
        grid's of NAME vn7.2 output and met-extractor in 2022."""

        assert idx1.nlevels == idx2.nlevels
        num_levels = idx1.nlevels

        # a stricter check is idx_i.equals(idx_0)

        levels_close = []
        for i in range(num_levels):
            close_i = allclose(idx1.get_level_values(i),idx2.get_level_values(i),atol=atol,rtol=0)
            levels_close += [close_i]

        return all(levels_close)

    def raster_to_series(self, raster_fn):

        with rio_open(raster_fn,'r') as host_raster:
            host_arr = host_raster.read(1)
            shape = host_raster.shape

            # determine coordinates
            coords = [host_raster.xy(i,j) for i in range(shape[0]) for j in range(shape[1])]
            lons = unique([ci[0] for ci in coords])
            lats = unique([ci[1] for ci in coords])
            assert shape == (lats.size,lons.size)

        # build into a dataframe
        # (rasters start in the top left, so descending latitude coordinates)
        host_df = DataFrame(data=host_arr,index=lats[::-1],columns=lons)
        host_df.index.name = 'latitude'
        host_df.columns.name = 'longitude'
        # rearrange to ascending latitude corodinates
        host_df.sort_index(axis='rows',inplace=True)
        # make spatial coordinates a multi-index, like for dep and env suit csvs
        host_series = host_df.stack()

        return host_series

    def rasters_to_csv(
            self,
            raster_fns_dict: dict,
            csv_fn: str,
            ):
        """Takes a dictionary of raster files with associated times and saves them
        as rows of a single csv. The csv columns and index structure matches model
        outputs as expected by the epi model. Used to prepare the host data."""

        host_serieses = []
        first = True
        for date_valid_from, raster_fn in raster_fns_dict.items():

            host_series = self.raster_to_series(raster_fn)

            # for now, provide a nominal date of validity to enable a time column
            # so far, using mapspam which is a static map, so time is irrelevant
            host_series.name = date_valid_from

            # conform indices (handle float differences)
            if first:
                idx_0 = host_series.index

            if not first:
                idx_i = host_series.index

                indices_are_close = self.are_indices_close(idx_0,idx_i)
                assert indices_are_close, (f"Coordinates of host_rasters do not match.\nFailed for {raster_fn}.")
                host_series.index = idx_0

            first = False

            host_serieses += [host_series]

        host_df = DataFrame(host_serieses)

        host_df.to_csv(csv_fn)

        return

    def get_model_divided_by_host_fraction(
                self,
                dfm,
                hostCSV,
                model_colns=None,
                **kwargs):
        """when model_infection pressure has units of [ha_infected/ha_cell]
        we want [ha_infected/ha_wheat] to compare with surveys
        (because surveys only sample the wheat covered landscape)
        so we must load the host raster and divide all model results by it.

        TODO: Instead of doing as post-processing in coordinator, best do it within
        the ews-epidemiology package.
        """

        print('Converting units of prediction from ha_infected/ha_cell to ha_infect/ha_wheat')

        # load host raster
        host_fn = hostCSV
        host_df = read_csv(host_fn,index_col=0,header=[0,1])
        host_df.columns = host_df.columns.set_levels([lvl.astype('float') for lvl in host_df.columns.levels])
        host_df.index = to_datetime(host_df.index,format='%Y%m%d%H%M')
        host_df = host_df.T

        # conform the structure with infection dataframe

        # conform indices (coordinates)
        host_df.index = host_df.index.reorder_levels(['longitude','latitude'])
        host_df.sort_index(level=['longitude','latitude'],ascending=[True,False],inplace=True)

        indices_are_close = self.are_indices_close(host_df.index,dfm.index)
        assert indices_are_close, ('Coordinates of model grid do not match host map.')
        host_df.index = dfm.index

        # conform columns (dates)
        column_end_dates = dfm.columns.map(lambda x: x[-12:])
        model_dates = to_datetime(column_end_dates, format='%Y%m%d%H%M') - datetime.timedelta(days=1)
        dfm2 = dfm.copy()
        dfm2.columns = model_dates

        # Set a host value for every model date, based forward-filling dated map to
        # next available date
        host_df_resampled = host_df.reindex(dfm2.columns,axis='columns',method='ffill')
        assert not host_df_resampled.isna().any().any(), ('Dates of host rasters do not cover all model dates')

        # new approach, take advantage of pandas broadcasting
        print('Applying unit conversion to all columns in output')
        dfm3 = dfm2.divide(host_df_resampled)
        # Handle cases of zero division
        dfm3[host_df_resampled<=0]=0

        # check for anomalously large values
        where_too_big = dfm3 > 1.00001
        if any(where_too_big):
            msg = 'ERROR: Unit conversion failed, host area seems to be smaller than predicted infection area in a cell'
            print(msg)
            raise Exception

        # clip any values that are above 1
        # (Expect this is not needed, but may help resolve float precision issues)
        dfm3.clip(0.,1.,inplace=True)

        # Retain original column names
        dfm3.columns = dfm.columns

        return dfm3

    def process_in_job_epi(self, jobPath,status,config,component):
        logger.info('started process_in_job_epi()')

        # TODO: Some of this is modifying config before epi model is run. Determine
        # how to account for that


        # If this job is continuing the results of a past epi run, there are
        # several additional requirements:
        
        # - The epi model will run from the reference date to the end date. The
        #   start date to reference date defines the range of the job that is being
        #   continued.

        # - Each model named under config['Epidemiology']['Epi'] will need a
        #   corresponding set of arguments for reading in as for depo and env, e.g.
        #   config['Epidemiology']['model_name'].

        # - depo and env file listers will be applied only for the reference date
        #   to end date, so choose an apporporiate lister, e.g.
        #   list_onefile_operational

        is_continue = config['Epidemiology'].get('continue',False)

        # initialise any needed variables

        reference_date_str = config['StartString']
        reference_date = datetime.datetime.strptime(reference_date_str,'%Y%m%d')

        start_date, end_date = calc_epi_date_range(reference_date_str,config['Epidemiology']['CalculationSpanDays'])

        start_string = start_date.strftime('%Y-%m-%d-%H%M')
        start_string_short = start_date.strftime('%Y%m%d%H%M')
        end_string = end_date.strftime('%Y-%m-%d-%H%M')

        # update config accordingly
        config['ReferenceTime'] = reference_date_str
        config['StartTime'] = start_string
        config['StartTimeShort'] = start_string_short
        config['EndTime'] = end_string

        # Used to get the last jobs to continue or update from
        yesterday_date = datetime.datetime.strptime(reference_date_str,'%Y%m%d') - datetime.timedelta(days=1)
        yesterday_string = yesterday_date.strftime('%Y%m%d')

        diseases = config['Epidemiology']['DiseaseNames']

        # get list of variable names to be loaded from deposition input
        depo_variable_names =  config['Epidemiology']['Deposition']['VariableNames']
        assert len(depo_variable_names) == len(diseases)

        # loop over each sub region

        region = config['RegionName']
        #for region in config['SubRegionNames']:

        for disease in diseases:

            assert disease in disease_latin_name_dict.keys()

            config['SubRegionName'] = region
            config['DiseaseName'] = disease

            config_epi = config['Epidemiology'].copy()

            # TODO: CAUTION: Any iterations (e.g. disease or sub-region) are hidden
            # in jobPath, and not retained in the config file. This is a provlem for
            # process_EWS_plotting_epi which receives a single config file and must
            # try a fudge to retrieve details for each iteration.
            # This should be improved, either by making the one config file
            # aware of all of the iterations, or looping over iterations in
            # Processor.py with one iteration-specific config.
            case_specific_path = f"{jobPath}/{region}/{disease}/"
            Path(case_specific_path).mkdir(parents=True, exist_ok=True)

            logger.info(f"Preparing for epidemiology calc of {disease} in {region}")

            # create config_filename to describe job configuration
            config_filename = self.create_epi_config_string(config,case_specific_path,start_string,end_string)

            # prepare a directory for input data
            jobDataPath = f"{case_specific_path}/input_data/"
            Path(jobDataPath).mkdir(parents=True, exist_ok=True)

            lastjobDataPath = jobDataPath.replace(f"EPI_{reference_date_str}",f"EPI_{yesterday_string}")

            # configure filename of prepared deposition data

            if 'Deposition' in config_epi:

                # determine which variable name to load for this disease
                disease_idx = [i for i,j in enumerate(diseases) if j==disease][0]

                variable_name = depo_variable_names[disease_idx]

                try:
                    prep.gather_dependent_models(
                            config_epi,
                            config,
                            variable_name,
                            start_date,
                            reference_date,
                            end_date,
                            jobDataPath,
                            lastjobDataPath,
                            component='Deposition')
                except:
                    
                    logger.exception(f"Unexpected error in {component} data preparation")
                    status.reset('ERROR')
                    end_job(status, premature=True)
            
            # configure filename of prepared environment data

            if 'Environment' in config_epi:

                logger.info('Preparing environmental suitability data')

                try:
                    prep.gather_dependent_models(
                            config_epi,
                            config,
                            variable_name,
                            start_date,
                            reference_date,
                            end_date,
                            jobDataPath,
                            lastjobDataPath,
                            component='Environment')
                
                except:

                    logger.exception(f"Unexpected error in {component} data preparation")
                    status.reset('ERROR')
                    end_job(status, premature=True)

        
            # prepare a copy of the host data

            logger.info('Preparing a copy of the host raster data')

            # TargetRaster defines the grid that the epi model works on.
            assert 'TargetRaster' in config_epi['Host']

            # It should have been generated in advance by the user, by reprojecting
            # the available host map (e.g. MapSPAM) to the NAME output grid.
            # wheat_raster_reprojection.py is available to support this.

            if 'HostRasters' in config_epi['Host']:
                # HostRasters is a dictionary with date: filename entries describing
                # different host rasters valid at different times i.e. a simple
                # representation of dynamic host, so prepare a host file as is done
                # for the Deposition and Environment components.

                # All host maps should have the same spatial grid as the TargetRaster

                rasters_dict = config_epi['Host']['HostRasters']

                dst_host_csv = f"{jobDataPath}/data_input_host.csv"

                self.rasters_to_csv(rasters_dict,dst_host_csv)

            else:
                # There is a host raster applicable to all times, i.e. static host

                src_host = config_epi['Host']['TargetRaster']
                fn_host = os.path.basename(src_host)
                dst_host = f"{jobDataPath}/{fn_host}"

                # copy the tif to the job directory and refer to that instead
                shutil.copyfile(src_host,dst_host)
                config_epi['Host']['TargetRaster'] = dst_host

                logger.info('Preparing a copy of the host data as csv')

                dst_host_csv = dst_host.replace('.tif','.csv')

                self.rasters_to_csv(
                        {'201001010000': dst_host},
                        dst_host_csv)

            config_epi['Host']['HostCSV'] = dst_host_csv
            config_epi['Host']['FileNamePrepared'] = dst_host_csv

            # Preparing any continue-run files
            if is_continue is True:

                logger.debug('This is a continue run.')

                for ci in config_epi['Epi']:
                
                    model_name = ci['model']

                    # Get results of last day and prepare as input

                    logger.info(f"Getting a copy of the {model_name} model data to continue from")
                    
                    try:
                        prep.gather_dependent_models(
                                config_epi,
                                config,
                                variable_name,
                                start_date,
                                reference_date,
                                end_date,
                                jobDataPath,
                                lastjobDataPath,
                                component=model_name)
                    except:
                        
                        logger.exception(f"Unexpected error in {model_name} data preparation")
                        status.reset('ERROR')
                        end_job(status, premature=True)


            # provide fundamental config elements to config_epi
            for k,v in config.items():
                if k not in short_name.keys():
                    config_epi[k]=v

            if is_continue is True:
                # This will be used in the epi model and not the data preparation
                continue_start_date = reference_date+datetime.timedelta(hours=3)
                config_epi['StartTime'] = continue_start_date.strftime('%Y-%m-%d-%H%M')
                config_epi['StartTimeShort'] = continue_start_date.strftime('%Y%m%d%H%M')

            def print_item(item):
                logger.debug(f"Item {item}")
                logger.debug(json.dumps(item,indent=2))
            def iterate(items):
                for item in items.items():
                    if hasattr(item,'items'):
                        # iterate
                        iterate(item)
                    else:
                        print_item(item)

            logger.debug('Incremental configuration looks like:')

            iterate(config_epi)

            logger.debug('Complete configuration looks like:')
            logger.debug(json.dumps(config_epi,indent=2))

            # write the complete configuration file to job directory
            with open(f"{case_specific_path}/{config_filename}.json",'w') as write_file:
                json.dump(config_epi,write_file,indent=4)

            # run epi model

            logger.info('About to run the epi model.')


            try:
                model.run_epi_model(f"{case_specific_path}/{config_filename}.json")
            except:
                logger.exception('Unexpected error in EpiModel')
                raise

            # perform calc on output

            def calc_total(arr):
                return 'total', arr.sum()

            def calc_max(arr):
                return 'maximum', arr.max()

            def calc_mean(arr):
                return 'mean', arr.mean()

            for epiconf in config_epi['Epi']:

                outfile = epiconf["infectionRasterFileName"]

                with rio_open(outfile+'.tif','r') as infectionRaster:
                    infection = infectionRaster.read(1)

                    # define function to quantify overall result, for easy check
                    # TODO: Create a more meaningful result?
                    # TODO: make this configurable
                    analysis_func = calc_mean

                    analysis_desc, analysis_value = analysis_func(infection)

                    logger.info(f"For case {outfile}")
                    logger.info('Infection {:s} is {:.2e}'.format( analysis_desc, analysis_value))

                    # to save tif as png for easy viewing
                    logger.debug('Saving tif output as png for easier viewing')
                    plotRaster.save_raster_as_png(outfile)

            # comparison figure

            # TODO: make this plot configurable? with function or args?
            #logger.info('Plotting epi output alongside contributing components')
            # figure_func = getattr(EpiAnalysis,'plot_compare_host_env_dep_infection')
            logger.info('Plotting composite image of epi formulations')
            figure_func = getattr(EpiAnalysis,'plot_compare_epi_cases')

            # isolate the config for this function, in case of modifications
            config_epi_for_comparison = config_epi.copy()

            fig,axes,cases = figure_func(
                    config_epi_for_comparison,
                    start_str = start_string,
                    end_str = end_string)

            SaveFileName = f"{case_specific_path}/EPI_{config_filename}_comparison"

            fig.savefig(SaveFileName+'.png',dpi=300)

            # slice the epi results into before forecast and in forecast

            for epiconf in config_epi['Epi']:

                outfile = epiconf["infectionRasterFileName"]+'_progression.csv'

                # load the full epi results
                df_full = read_csv(outfile,header=[0],index_col=[0,1])
                column_date_fmt = f"{config_epi['StartTimeShort']}_%Y%m%d%H%M"
                df_full_dates = to_datetime(df_full.columns.astype('str'),format=column_date_fmt)

                unit_description = ''

                # convert units from ha_infected/ha_cell to ha_infected/ha_wheat

                if epiconf['rescale_output_by_host_raster'] is True:

                    unit_description = '_per_ha_wheat'

                    model_colns = df_full.columns

                    df_full = self.get_model_divided_by_host_fraction(
                            df_full,
                            config_epi['Host']['HostCSV'],
                            model_colns = model_colns)

                    # save to csv
                    outfile_hawheat = f"{epiconf['infectionRasterFileName']}{unit_description}_progression.csv"
                    df_full.to_csv(outfile_hawheat,header=True,index=True)

                    outfile_hawheat_final = f"{epiconf['infectionRasterFileName']}{unit_description}.csv"
                    df_full.iloc[:,-1].to_csv(outfile_hawheat_final,header=True,index=True)

                # determine date to cut with
                # plus 1 minute so midnight is associated with preceding day
                date_to_cut = datetime.datetime.strptime(config_epi['StartString']+'0001','%Y%m%d%H%M')
                dates_after_cut = df_full_dates >= date_to_cut
                idx = argmax(dates_after_cut)-1

                if idx == -1:
                    # only working on forecast dates so no season so far, skip
                    continue

                # build seasonsofar dataframe (only need the last date)
                df_seasonsofar = df_full.iloc[:,idx]

                # check column name is defined as expected
                # from epi start time to forecast start time
                column_name = f"{config['StartTimeShort']}_{config['StartString']}0000"
                assert df_seasonsofar.name == column_name

                #  save to csv
                fn_seasonsofar = f"{epiconf['infectionRasterFileName']}{unit_description}_seasonsofar.csv"
                df_seasonsofar.to_csv(fn_seasonsofar,header=True,index=True)

                # build weekahead dataframe and save to csv
                df_fc_start = df_full.iloc[:,idx]
                df_fc_start_name = df_fc_start.name.split('_')[-1]

                df_fc_end = df_full.iloc[:,-1]
                df_fc_end_name = df_fc_end.name.split('_')[-1]

                df_weekahead = df_fc_end - df_fc_start

                # defined column name
                fn_weekahead = f"{epiconf['infectionRasterFileName']}{unit_description}_weekahead.csv"
                df_weekahead.name = '_'.join([df_fc_start_name,df_fc_end_name])

                # save to csv
                df_weekahead.to_csv(fn_weekahead,header=True,index=True)

        proc_out = {}
        # Output files available for upload
        proc_out['output'] = None
        # Processing files available for clearing
        proc_out['clearup'] = None

        return proc_out

    def process_EWS_plotting_epi(self, jobPath,config) -> [str]:
        '''Returns a list of output files for transfer.'''

        logger.info('started process_EWS_plotting_epi()')

        # initalise necessary variables from config

        start_date, end_date = calc_epi_date_range(config['StartString'],config['Epidemiology']['CalculationSpanDays'])

        start_string = start_date.strftime('%Y%m%d')
        end_string = end_date.strftime('%Y%m%d')

        epi_case_operational = config['Epidemiology']['EWS-Plotting']['EpiCase']

        if epi_case_operational == 'none':
            logger.info('Config specifies not to call to EWS-Plotting')
            return []

        diseases = config['Epidemiology']['DiseaseNames']

        # initialise environment
        sys_config = config['Epidemiology']['EWS-Plotting']['SysConfig']

        chart_config = config['Epidemiology']['EWS-Plotting']['ChartConfig']

        # use the first matching epi formulation
        # TODO: Is there a more efficient way to select?
        epi_filename = [ce['infectionRasterFileName'] for ce in config['Epidemiology']['Epi'] if ce['model']==epi_case_operational][0]

        dep_regionnames = ['SouthAsia','Ethiopia']

        # TODO get deposition_dir from config['Epidemiology']['Deposition']['PathTemplate']
        dep_regionname = 'Ethiopia' #SouthAsia

        deposition_dir = f"{config['WorkspacePath']}DEPOSITION_{start_string}/WR_NAME_{dep_regionname}_{start_string}/"

        # TODO: handle multiple diseases and regions in Processor as a loop, or in the config
        deposition_disease_name = [disease_latin_name_dict[disease]+'_DEPOSITION' for disease in diseases][0]

        ews_plot_dir = f"{jobPath}/plotting/"

        Path(ews_plot_dir).mkdir(parents=True, exist_ok=True)

        # loop over diseases
        ews_plotting_output_globs = []
        for disease in diseases:
            disease_short = disease.lower().replace('rust','')

            # a fudge, guess disease type
            # because config['Epidemiology']['ProcessInJob'] handles disease loop internally
            # assumes disease name is the last directory before the filename
            # TODO: handle multiple diseases and regions in Processor as a loop, or in the config
            disease_to_drop = os.path.dirname(epi_filename).split('/')[-1].replace('Rust','')
            disease_to_add = disease.replace('Rust','')
            epi_filename = epi_filename.replace(disease_to_drop,disease_to_add)

            map_title = "Integrated prediction of Wheat $\\bf{" + disease_to_add + "}$ Rust infection"
            if 'PlottingRegionName' not in config['Epidemiology']['EWS-Plotting']:
                plotting_region_name_lower = config['RegionName'].lower()
            else:
                plotting_region_name_lower = config['Epidemiology']['EWS-Plotting']['PlottingRegionName'].lower()

            epi_seasonsofar_fn = epi_filename+'_per_ha_wheat_seasonsofar.csv'

            epi_seasonincforecast_fn = epi_filename+'_per_ha_wheat.csv'

            seasonsofar_run_config = config['Epidemiology']['EWS-Plotting'].get('RunConfig_seasonsofar',None)

            # only run season so far (i.e. historic dates) if they exist
            if (seasonsofar_run_config is not None) & os.path.exists(epi_seasonsofar_fn):

                logger.info(f"Running EWS-Plotting with the following configs:\n{sys_config}\n{seasonsofar_run_config}\n{chart_config}")

                epi_processor_1 = EPIPostPostProcessor()
                epi_processor_1.set_param_config_files(sys_params_file_arg=sys_config,
                                                chart_params_file_arg=chart_config,
                                                run_params_file_arg=seasonsofar_run_config,
                                                epi_input_csv_arg=epi_seasonsofar_fn,
                                                disease_type_arg=disease_short+'_seasontodate',
                                                issue_date_arg=start_string,
                                                output_dir_arg=ews_plot_dir,
                                                wheat_sources_dir_arg=deposition_dir,
                                                wheat_source_disease_name_arg=deposition_disease_name,
                                                map_title_arg=map_title,
                                                chart_area_prefix=plotting_region_name_lower)
                epi_processor_1.process()

            # prepare command for seasonplusforecast

            run_config = config['Epidemiology']['EWS-Plotting']['RunConfig_seasonplusforecast']

            logger.info(f"Running EWS-Plotting with the following configs:\n{sys_config}\n{run_config}\n{chart_config}")

            epi_processor_2 = EPIPostPostProcessor()
            epi_processor_2.set_param_config_files(sys_params_file_arg=sys_config,
                                            chart_params_file_arg=chart_config,
                                            run_params_file_arg=run_config,
                                            epi_input_csv_arg=epi_seasonincforecast_fn, # for seasonplusforecast
                                            #epi_input_csv_arg=epi_filename+'_weekahead.csv', # for weekahead
                                            disease_type_arg=disease_short+'_seasonincforecast',
                                            issue_date_arg=start_string,
                                            output_dir_arg=ews_plot_dir,
                                            wheat_sources_dir_arg=deposition_dir,
                                            wheat_source_disease_name_arg=deposition_disease_name,
                                            map_title_arg=map_title,
                                            chart_area_prefix=plotting_region_name_lower)
            epi_processor_2.process()

            # check the output
            ews_plotting_output_dir = f"{ews_plot_dir}/images/"
            # TODO: Make this smarter, connected to the results of EWSPlottingEPIBase.plot_epi()
            ews_plotting_output_globs: [str] = []
            ews_plotting_output_globs += [f"{ews_plotting_output_dir}infection_{plotting_region_name_lower}_*{disease_short}*.png"]

            ews_plotting_output_globs = get_only_existing_globs(ews_plotting_output_globs,inplace=False)

            # check there is some output from EWS-plotting
            if not ews_plotting_output_globs:
                logger.error('EWS-Plotting did not produce any output')
                raise RuntimeError

        # provide to list for transfer
        ews_plotting_outputs: [str] = [item for ews_plotting_output in ews_plotting_output_globs for item in glob(ews_plotting_output)]

        return ews_plotting_outputs


if __name__ == '__main__':
    processor = ProcessorEpidemiology()
    processor.run_processor("Epidemiology", "EPI")
