#ProcessorSurveys.py
'''Functions to process the survey component.

Group surveys and run source gen on each groups:
Grouping is only run if it is defined in the config (otherwise it use all available surveys).
Groups need to have a name, a key to at least one column and a list of unique elements in that column to group by.
Example:
        "Groups" : {
            "PROD" : {
                "Origin" : ["CSV_01", "CSV_02", "ODK_01", "ODK_02"],
                "PublishedLevel" : ["Unpubl", "Publ", "Unknown"]
            },
            "WRT-Publ" : {
                "Origin" : ["CSV_01", "CSV_02", "ODK_01", "ODK_02"],
                "PublishedLevel" : ["Publ"]
            }
        },
        "GroupBy" : ["Origin", "PublishedLevel"],
        "GroupsToIgnore" : ["CSV_01", "CSV_02", "ODK_01", "ODK_02", "Unpubl", "Publ"], 

Alongside the defined groups, individual groups get created based on the unique elements in the columns defined in "GroupBy".
Groups can be ignored by defining "GroupsToIgnore" in the config.
'''

import csv
import datetime
import json
import logging
from glob import glob
from pathlib import Path
from shutil import copyfile

from numpy import all as np_all
from numpy import any as np_any
from pandas import read_csv, concat

from ews.coordinator.survey_servers.processor_surveys_odk import get_ODK_form_as_csv
from ews.coordinator.survey_servers.processor_surveys_odk_sa import get_ODK_SA_form_as_csv
from ews.coordinator.survey_servers.processor_surveys_odk_v2 import get_ODKv2_form_as_csv
from ews.coordinator.survey_servers.processor_surveys_wrsis import get_WRSIS_form_as_csv
from ews.coordinator.survey_servers.processor_surveys_wrt import get_WRT_form_as_csv
from ews.coordinator.survey_servers.processor_surveys_kobotoolbox import get_kobotoolbox_form_as_csv
from ews.coordinator.survey_servers.processor_surveys_new_odk import get_newODK_form_as_csv
from ews.coordinator.survey_servers.processor_surveys_new_odk2 import get_newODK2_form_as_csv
from ews.coordinator.processor_base import ProcessorBase
from ews.source_gen.clustering import run_case


logger = logging.getLogger(__name__)

class ProcessorSurveys(ProcessorBase):

    def process_pre_job(self, args):
        # return self.process_pre_job_survey(args)
        return True


    def process_in_job(self, jobPath, status, configjson, component) -> object:
        return self.process_in_job_survey(jobPath, status, configjson, component)


    def process_post_job(self, jobPath, configjson) -> [str]:
        return self.process_EWS_plotting_survey(jobPath, configjson)


    def __init__(self) -> None:
        super().__init__()

        self.GET_FORM_AS_CSV_DICT = {
            'ODK':         get_ODK_form_as_csv,
            'ODK_SA':      get_ODK_SA_form_as_csv,
            'kobotoolbox': get_kobotoolbox_form_as_csv,
            'WRSIS':       get_WRSIS_form_as_csv,
            'WRT':         get_WRT_form_as_csv,
            'ODKv2':       get_ODKv2_form_as_csv,
            'newODK':      get_newODK_form_as_csv,
            'newODK2' :    get_newODK2_form_as_csv,
        }

    def process_in_job_survey(self, jobPath,status,config,component):

        logger.info('started process_in_job_survey()')

        logger.debug('Performing download(s) from ODK server')

        credentials_filename = config['Survey']['ServerCredentialsFile']
        with open(credentials_filename) as credentials_file:

            cred: dict = json.load(credentials_file)

            assert 'forms' in cred.keys()

        csv_filenames = {}
        for form in cred['forms']:

            logger.debug(f"Starting to download {form['form_id']}")

            assert form['type'] in self.GET_FORM_AS_CSV_DICT

            func_get_form_as_csv = self.GET_FORM_AS_CSV_DICT[form['type']]

            csv_filename = func_get_form_as_csv(form, jobPath, config, status)

            csv_filenames[form['form_id']] = csv_filename

        # load each file of surveys as a dataframe
        forms = {}
        for form_name,form_fn in csv_filenames.items():

            # some define column types, hardwired for now
            col_types = {'comment':'str','KEY':'str'}

            form_df = read_csv(form_fn,dtype=col_types)

            forms[form_name] = form_df

        # create some standard dataframe modification functions
        def add_column(df,coln,value):
            df[coln]=value
            return

        def remove_column(df,coln,value):
            del df[coln]
            return

        def replace_column(df,coln,value):
            df[coln]=value
            return

        def filter_by_column(df,coln,value):
            # CAUTION: This requires surveyor to provide the correct country
            df.drop(df.loc[df[coln]!=value].index,inplace=True)
            #TODO : for Kenya data, provide a coordinate-based filter
            return

        def filter_by_list(df,coln,values):
            # CAUTION: This requires surveyor to provide the correct list of countries
            df.drop(df.loc[~df[coln].isin(values)].index,inplace=True)
            return

        func_types = {
            'add': add_column,
            'remove' : remove_column,
            'replace' : replace_column,
            'filter' : filter_by_column,
            'filter_by_list' : filter_by_list
        }

        # simple format alignment using edits on config
        # (should this need to be much more sophisticated, reconsider the workflow)
        if 'FormEdits' in config['Survey']:

            form_edits = config['Survey']['FormEdits']

            # loop over each form
            for form_name, edits in form_edits.items():

                form_df = forms[form_name]

                # loop over each type of edit
                for func_type, columns in edits.items():

                    # check the function is available
                    assert func_type in func_types

                    # loop over each column to modify
                    for coln,val in columns.items():

                        # apply the edit
                        func_types[func_type](form_df,coln,val)

        # Merge additional SurveyData files and rearrange columns to be consistent
        # Assumes that the same columns are present in all forms
        # and that the first form is the standard

        first=True
        for dfi in forms.values():

            if first:
                standard_columns = dfi.columns.tolist()
                dfm = dfi

                logger.debug(f"First processed form contains {dfm.shape[0]} records")

                first=False
                continue

            # re-order columns to match first case (presumed standard format)
            # and fill missing columns with empty strings
            dfi = dfi.reindex(standard_columns,fill_value='',axis='columns')

            logger.debug(f"Next processed form contains {dfi.shape[0]} records")

            dfm = concat([dfm,dfi],axis='rows')

        # save the result
        Export_csv_path = f"{jobPath}/ExportCSV/"
        Path(Export_csv_path).mkdir(parents = True, exist_ok = True)
        forms_fn = f"{Export_csv_path}/Merged_SurveyData.csv"
        dfm.to_csv(forms_fn,index=False,quoting=csv.QUOTE_MINIMAL)

        logger.debug(f"Preparing to apply removals and additions to survey data")

        processed_surveys_filepath = f"{Export_csv_path}/Processed_SurveyData.csv"

        survey_errors_to_remove_filepath = f"{config['ResourcesPath']}/coordinator/assets/SURVEYDATA_MANUAL/SurveyDataErrorsToRemove.csv"
        survey_additions_filepath = f"{config['ResourcesPath']}/coordinator/assets/SURVEYDATA_MANUAL/LIVE_SURVEYDATA_TOUSE.csv"

        # perform here in python, using the 'KEY' column
        # check the key column is unique, if not raise a warning and remove duplicates

        if dfm['KEY'].unique().size != dfm['KEY'].size:
            status.reset('WARNING')
            logger.warning(f"KEY column is not unique, removing duplicates")
            # count the number of duplicates
            n_duplicates = dfm.shape[0] - dfm['KEY'].unique().size
            # drop the duplicates
            dfm = dfm.drop_duplicates(keep='first')
            logger.warning(f"Removed {n_duplicates} duplicates")

        df_rm = read_csv(survey_errors_to_remove_filepath,dtype='str')
        keys_to_rm = df_rm['KEY']

        # check that all of the keys to remove exist in the original data
        rm_keys_found = df_rm['KEY'].isin(dfm['KEY'])
        n_rm_keys_found = rm_keys_found.sum()
        n_rm_keys = rm_keys_found.size
        if not np_all(rm_keys_found):
            # this might happen if the run date is in the past
            logger.warning(f"Only found {n_rm_keys_found} of {n_rm_keys} survey errors to remove")

            rm_keys_not_found = df_rm[~rm_keys_found]
            logger.debug(f"Erroneous entries not found are:\n{rm_keys_not_found}")

            logger.debug(f"Type of keys that can be found include:\n{dfm['KEY'].dtype}")

            dfm_short_keys = [val for val in dfm['KEY'].values if len(str(val)) <10]
            logger.debug(f"Keys that can be found include:\n{dfm_short_keys}")

        # identify which surveys to remove
        idx_to_rm = dfm['KEY'].apply(lambda cell: cell in keys_to_rm.values)

        #drop them in-place
        dfm = dfm[~idx_to_rm]
        logger.info(f"Removed {n_rm_keys_found} erroneous surveys")

        # add the extra entries
        df_add = read_csv(survey_additions_filepath,dtype='str')
        n_add_keys = df_add.shape[0]
        df_join = concat([dfm,df_add])
        assert dfm.shape[0]+df_add.shape[0] == df_join.shape[0], 'Unexpected result of including additional surveys'

        logger.info(f"Added {n_add_keys} additional surveys")

        # save as processed
        df_join.to_csv(processed_surveys_filepath,index=False,quoting=csv.QUOTE_MINIMAL)

        logger.debug('Preparing clustering calculation')

        date = datetime.datetime.now()

        # prepare environment for clustering calc

        upload_directory = f"{jobPath}/upload"
        Path(upload_directory).mkdir(parents=True, exist_ok=True)

        if 'Groups' in config['Survey']:
            # if 'Groups' is defined in the config, create grouped survey files and run python version

            logger.info('Preparing grouped survey files')
            group_directory = f"{jobPath}/Groups"
            Path(group_directory).mkdir(parents=True, exist_ok=True)

            # creating initial groups
            groups = config['Survey']['Groups']

            # check if columns requested in GroupBy are present in the dataframe
            assert all([group_by in df_join.columns for group_by in config['Survey']['GroupBy']]), 'Column(s) requested in GroupBy are not present in the dataframe'

            for group_by in config['Survey']['GroupBy']:
                logger.debug(f"grouping by {group_by}")

                # handle NaN values
                if df_join[group_by].isna().any():
                    logger.warning(f"Grouping by {group_by} contains NaN values. Filling NaN values with 'Unknown'")
                    df_join[group_by] = df_join[group_by].fillna('Unknown')
                
                groups_list = df_join[group_by].unique()
                # remove 'Unknown' from the list of groups
                groups_list = [i for i in groups_list if i != 'Unknown']

                assert not np_any([i in groups_list for i in groups]), f"Group name {[i for i in groups_list if i in groups]} already present in the {group_by} column. Please rename the group name in the config file."
                
                groups.update({i:{group_by:[i]} for i in groups_list})

            # remove groups that are listed in GroupsToIgnore
            if 'GroupsToIgnore' in config['Survey']:
                groups_to_ignore = config['Survey']['GroupsToIgnore']

                # add groups to ignore if all elements of the group are in the list of groups to ignore
                for group_name in groups:
                    if '+' in group_name:
                        group_name_separated = group_name.split('+')
                        if all([i in groups_to_ignore for i in group_name_separated]):
                            logger.debug(f"Adding group {group_name} to list of groups to ignore")
                            groups_to_ignore.append(group_name) # TODO: rename group based on the unique element

                for group_name in groups_to_ignore:
                    if group_name in groups:
                        logger.debug(f"Removing group {group_name} from list of groups")
                        del groups[group_name]

            for group_name,group_content in groups.items():

                logger.info(f"Creating survey group {group_name} which includes {group_content}")

                # applying the grouping, keeping only the surveys that have the elements listed in the group
                df_group = df_join
                for group_by,group_elements in group_content.items():
                    df_group = df_group.loc[df_group[group_by].isin(group_elements)]

                group_surveys_filename = f"surveys_{group_name}.csv"
                group_surveys_filepath = f"{group_directory}/{group_surveys_filename}"

                df_group.to_csv(group_surveys_filepath, index=False, quoting=csv.QUOTE_MINIMAL)

                output_directory = f"{jobPath}/source_gen/{group_name}"
                Path(output_directory).mkdir(parents=True, exist_ok=True)

                if 'SourcesConfigs' in config['Survey'] and group_name in config['Survey']['SourcesConfigs']:
                    logger.info(f"Running source gen for {group_name} group wih config {config['Survey']['SourcesConfigs'][group_name]}")
                    sources_config = config['Survey']['SourcesConfigs'][group_name]
                else:
                    logger.info(f"Running source gen for {group_name} group wih default config {config['Survey']['SourcesConfigDefault']}")
                    sources_config = config['Survey']['SourcesConfigDefault']
                
                sources_path = run_case(
                        config_path = sources_config,
                        survey_path = group_surveys_filepath,
                        survey_format = config['Survey']['SurveyFormat'],
                        surveyor_name_col = config['Survey']['SurveyorNameCol'],
                        reference_date = config['StartString'],
                        # Day 0 (current day) is always included
                        # Days -2,-1 and 0 are needed to initialise spores in NAME
                        # days 1-7 are forecast days
                        # days 8 and 9 are buffers in case NAME needs to run with the
                        # previous day's source job
                        day_offsets = [-2,9],
                        output_dir = output_directory)

                logger.debug('Placing copy of result in job directory with conventional name')

                output_filename = f"sources_{group_name}_{config['StartString']}.csv"
                output_path = f"{jobPath}/upload/{output_filename}"

                logger.debug(f"as {output_path}")

                copyfile(sources_path, output_path)

                # THIS CAN BE REMOVED ONCE THE GROUPS ARE PROPERLY PICKED UP BY THE METOFFICE
                if (group_name == 'PROD'):
                    logger.debug('Additionally placing copy of PROD result in job directory without group name')

                    output_filename = f"sources_{config['StartString']}.csv"
                    output_path = f"{jobPath}/upload/{output_filename}"

                    logger.debug(f"as {output_path}")

                    copyfile(sources_path, output_path)
        else:
            # run python version without grouping surveys

            output_directory = f"{jobPath}/source_gen"
            Path(output_directory).mkdir(parents=True, exist_ok=True)

            sources_path = run_case(
                    config_path = config['Survey']['SourcesConfigDefault'],
                    survey_path = processed_surveys_filepath,
                    survey_format = config['Survey']['SurveyFormat'],
                    surveyor_name_col = config['Survey']['SurveyorNameCol'],
                    reference_date = config['StartString'],
                    # Day 0 (current day) is always included
                    # Days -2,-1 and 0 are needed to initialise spores in NAME
                    # days 1-7 are forecast days
                    # days 8 and 9 are buffers in case NAME needs to run with the
                    # previous day's source job
                    day_offsets = [-2,9],
                    output_dir = output_directory)

            logger.debug('Placing copy of result in job directory with conventional name')

            output_filename = f"sources_{config['StartString']}.csv"
            output_path = f"{jobPath}/upload/{output_filename}"

            logger.debug(f"as {output_path}")

            copyfile(sources_path, output_path)


        upload_filenames = f"sources_*{config['StartString']}.csv"
        upload_path = f"{jobPath}/upload/{upload_filenames}"

        # glob list of output files
        upload_path_list = glob(upload_path)

        proc_out = {}
        # Output files available for upload
        proc_out['output'] = upload_path_list
        # Processing files available for clearing
        proc_out['clearup'] = None

        return proc_out

    #TODO
    def process_EWS_plotting_survey(self, jobPath,config):
        '''Returns a list of output files for transfer.'''

        logger.info('started process_EWS_plotting_survey(), nothing to do')

        pass
        return []


if __name__ == '__main__':
    print(__name__)
    processor = ProcessorSurveys()
    processor.run_processor("Survey", "SURVEYDATA")
